package DinnerParty;

import org.xml.sax.*;
import uk.co.wilson.xml.MinML;

import java.net.*;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

public class DinnerParty extends MinML {
  private URL datapath;
  private String storiesName;
  //VALUE = String[3]
  // [0] = Story control file
  // [1] = Story title
  // [2] = Story about description
  private Vector theStories = new Vector();
  // [0] = open delimiter
  // [1] = close delimter
  // [2] = choice delimiter
  private String [] keyfile = new String[3];
  //VALUE = String[2]
  // [0] = template file
  // [1] = title
  private Vector theTemplates = new Vector();
  //KEY = Variable name
  //VALUE = String[].
  //        String[0]=Description
  //        String[1]=Current value
  //        String[2..n] = Choices (only present if a choice variable type)
  private Hashtable theVars = new Hashtable();
  //Convenience because the display interfaces like integers.
  private Vector varKeys = new Vector();

  //A places to build up information
  private String [] aStory=null;
  private String [] aTemplate=null;
  private String aVarName=null;
  private String aVarDesc=null;
  private String anOption=null;
  private Vector aVar = new Vector();
  //A place to hold the current element index during XML parse.
  private int thisElement = 0;
  private int theParent;  //The major type of object (STORY, TEMPLATE, STRING, SELECT)

  private static final String [] ELEMENTS = {"TITLE", "DESCRIPTION", "STRING",
    "CHOICE", "OPTION", "STORY", "TEMPLATE", "DELIMITERS", "ABOUT", "STORYLIST", "TEMPLATELIST"};
  private static final int TITLE = 0;
  private static final int DESCRIPTION = 1;
  private static final int STRING = 2;
  private static final int CHOICE = 3;
  private static final int OPTION = 4;
  private static final int STORY = 5;
  private static final int TEMPLATE = 6;
  private static final int DELIMITERS = 7;
  private static final int ABOUT = 8;
  private static final int STORYLIST = 9;
  private static final int TEMPLATELIST = 10;

  public DinnerParty() {
  }
  public DinnerParty(URL path, String file) {
    setFilesPath(path);
    setStoriesFile(file);
  }
  public void setFilesPath(URL path) {
    datapath = path;
  }
  public void setStoriesFile(String file) {
    storiesName = file;
  }
  public int loadStories() {
    URL storiesURL;
    BufferedReader storyReader;

    theStories.removeAllElements(); //Reset to empty

    try {
      storiesURL = new URL(datapath, storiesName);
      storyReader =
      new BufferedReader(new InputStreamReader(storiesURL.openStream()));
      parse(storyReader);
      storyReader.close();
      return theStories.size();
    }
    catch (MalformedURLException ex) {ex.printStackTrace();}
    catch (IOException ex) {ex.printStackTrace();}
    catch (SAXException ex) {ex.printStackTrace();}

    return 0; //Couldn't even start reading....
  }
  public int getStoryCount() {
    return theStories.size();
  }
  public String getStoryTitle(int s) {
    String [] aStory = (String [])theStories.elementAt(s);
    return aStory[1];
  }
  public String getStoryAbout(int s) {
    String [] aStory = (String [])theStories.elementAt(s);
    return aStory[2];
  }
  public void loadStoryList(int s) {
    URL storyURL;
    BufferedReader storyReader;

    String [] aStory = (String [])theStories.elementAt(s);
    theTemplates.removeAllElements();
    theVars.clear();
    varKeys.removeAllElements();

    try {
      storyURL = new URL(datapath, aStory[0]);
      storyReader =
        new BufferedReader(new InputStreamReader(storyURL.openStream()));
      parse(storyReader);
      storyReader.close();
    }
    catch (MalformedURLException ex) {ex.printStackTrace();}
    catch (IOException ex) {ex.printStackTrace();}
    catch (SAXException ex) {ex.printStackTrace();}
  }
  public int getVariableCount() {
    return theVars.size();
  }
  public String getVariableName(int v) {
    return (String) varKeys.elementAt(v);
  }
  public String getVariableType(int v) {
    String [] aVar = (String [])theVars.get(varKeys.elementAt(v));
    if (aVar.length>2) {
      return "choice";
    }
    return "string";
  }
  public String getVariableHelp(int v) {
    return ((String [])theVars.get(varKeys.elementAt(v)))[0];
  }
  public void setVariableValue(int v, String s) {
    String [] aVar = (String [])theVars.get(varKeys.elementAt(v));
    aVar[1] = s;
  }
  public String [] getVariableChoice(int v) {
    String [] aVar = (String [])theVars.get(varKeys.elementAt(v));
    String [] choices = new String[aVar.length-2];
    for (int i=0; i<choices.length; i++) {
      choices[i] = aVar[i+2];
    }
    return choices;
  }
  public String [] getTemplateTitles() {
    String [] titles = new String[theTemplates.size()];
    String [] aFile;
    for (int i=0; i<titles.length; i++) {
      aFile = (String [])theTemplates.elementAt(i);
      titles[i] = aFile[1];
    }
    return titles;
  }
  public String fillTemplate(int t) {
    int i,ii;
    int e,ee;
    String tline;
    String aVar;
    String cVar;
    String [] values;
    Vector choices;
    StringBuffer result = new StringBuffer(8192);

    String [] aFile = (String [])theTemplates.elementAt(t);
    try {
      URL templateURL = new URL(datapath, aFile[0]);
      BufferedReader templateReader =
        new BufferedReader(new InputStreamReader(templateURL.openStream()));
      while ((tline=templateReader.readLine())!=null) {
        if (tline.length()>0) {
          i=0;
          while ((e = tline.indexOf(keyfile[0],i)) != -1) {
            if (e != i) {    //Save the part up to the delimiter
              result.append(tline.substring(i,e));
            }
            i = e + keyfile[0].length();
            e = tline.indexOf(keyfile[1],i);
            if (e != -1) {
              aVar = tline.substring(i,e);
              //System.out.println("i="+i+" e="+e+" v="+aVar);
              i = e + keyfile[1].length();
              //First try the easy one....
              if (theVars.containsKey(aVar)) {
                values = (String [])theVars.get(aVar);
                result.append(values[1]);
              } else {
                //Must have a choice.. so it's more complicated...
                choices = new Vector();
                ii=0;
                ee = aVar.indexOf(keyfile[2], ii);
                if (ee != -1 ) {
                  cVar = aVar.substring(ii, ee);
                  ii = ee + keyfile[2].length();
                  while ((ee=aVar.indexOf(keyfile[2],ii)) != -1) {
                    choices.addElement(aVar.substring(ii,ee));
                    ii = ee + keyfile[2].length();
                  }
                  if (ii<aVar.length()) {
                    choices.addElement(aVar.substring(ii, aVar.length()));
                  }
                  if (theVars.containsKey(cVar)) {
                    values = (String [])theVars.get(cVar);
                    ee = Integer.parseInt(values[1]);
                    if (ee < choices.size()) {
                      result.append((String)choices.elementAt(ee));
                    } else {
                      System.out.println("Choice is invalid:"+aVar);
                      result.append(keyfile[0]);
                    }
                  } else {
                    System.out.println("Unknown variable:"+aVar);
                    result.append(keyfile[0]);
                    result.append(aVar);
                    result.append(keyfile[1]);
                  }
                } else {
                  System.out.println("Unknown variable:"+aVar);
                  result.append(keyfile[0]);
                  result.append(aVar);
                  result.append(keyfile[1]);
                }
              }
            } else {
              System.out.println("Missing suffix:"+tline);
              result.append(keyfile[0]);  //Add the prefix back in and proceed.
            }
          }
          //Add the remainder if there is some left.
          if (i < tline.length()) {
            result.append(tline.substring(i,tline.length()));
          }
        }
        result.append("\n");
      }
    }
    catch (MalformedURLException ex) {ex.printStackTrace();}
    catch (IOException ex) {ex.printStackTrace();}
    //System.out.println(result.toString());
    return result.toString();
  }
  public void startDocument() {
    //Nothing of interest here
    //System.out.println("Start of Document");
  }

  public void endDocument() {
    //Nothing of interest here
    //System.out.println("End of Document");
  }

  public void startElement (String name, AttributeList attributes) {
    int i;

    //System.out.println("Start of Element: \"" + name + "\"");
    //System.out.println("Start of Attributes");

    //for (i = 0; i < attributes.getLength(); i++) {
      //System.out.println("Name: \"" + attributes.getName(i) + "\" Type: " + attributes.getType(i) + " Value: \"" + attributes.getValue(i) + "\"");
    //}

    //System.out.println("End of Attributes");

    for (i=0; i<ELEMENTS.length; i++) {
      if (name.equalsIgnoreCase(ELEMENTS[i])) {
        thisElement = i;
        break;
      }
    }
    switch (thisElement) {
      case STORY:           //Only interested in STORY here...
        aStory = new String [3];
        aStory[0] = attributes.getValue(0);
        aStory[1] = "";
        aStory[2] = "";
        theParent = STORY;
        break;
      case DELIMITERS:
        for (i=0; i<attributes.getLength(); i++) {
          if (attributes.getName(i).equalsIgnoreCase("open")) {
            keyfile[0]=attributes.getValue(i);
          } else if (attributes.getName(i).equalsIgnoreCase("close")) {
            keyfile[1]=attributes.getValue(i);
          } else if (attributes.getName(i).equalsIgnoreCase("choice")) {
            keyfile[2]=attributes.getValue(i);
          }
        }
        break;
      case TEMPLATE:
        aTemplate = new String[2];
        aTemplate[0]=attributes.getValue(0);
        aTemplate[1]="";
        theParent = TEMPLATE;
        break;
      case CHOICE:
      case STRING:
        //STRING & CHOICE actually start out the same...
        aVarName = attributes.getValue(0);
        aVar.removeAllElements();
        aVarDesc = "";
        theParent = thisElement;
        break;
      case OPTION:
        anOption="";
        break;
    }
  }

  public void endElement (String name) {
    //System.out.println("End of Element: \"" + name + "\"");
    for (int i=0; i<ELEMENTS.length; i++) {
      if (name.equalsIgnoreCase(ELEMENTS[i])) {
        thisElement = i;
        break;
      }
    }
    switch (thisElement) {
      case OPTION:
        aVar.addElement(anOption);
      break;
      case STORY:
        theStories.addElement(aStory);
      break;
      case TEMPLATE:
        theTemplates.addElement(aTemplate);
      break;
      case STRING:
        String [] tmp = new String[2];
        tmp[0]=aVarDesc;
        tmp[1]="";
        theVars.put(aVarName, tmp);
        varKeys.addElement(aVarName);
      break;
      case CHOICE:
        aVar.insertElementAt("0",0);
        aVar.insertElementAt(aVarDesc,0);
        tmp = new String[aVar.size()];
        aVar.copyInto(tmp);
        theVars.put(aVarName,tmp);
        varKeys.addElement(aVarName);
      break;
    }
    thisElement = -1; //So we ignore stuff outside the elements...
  }

  public void characters (char ch[], int start, int length) {
    //System.out.println("Characters: \"" + new String(ch, start, length) + "\"");
    switch (thisElement) {
      case TITLE:
        switch (theParent) {
          case STORY:
            aStory[1] += new String(ch, start, length);
            break;
          case TEMPLATE:
            aTemplate[1] += new String(ch, start, length);
            break;
        }
      break;
      case DESCRIPTION:
        switch (theParent) {
          case STORY:
            aStory[2] += new String(ch, start, length);
            break;
          case STRING:
          case CHOICE:
            aVarDesc += new String(ch, start, length);
            break;
        }
      break;
      case ABOUT:
        aStory[2] += new String(ch, start, length);
        break;
      case OPTION:
        anOption += new String(ch, start, length);
        break;
    }
  }

  public void fatalError (SAXParseException e) throws SAXException {
    System.out.println("Error: " + e);
    throw e;
  }
}
